/*************************************************************
 * File  :  license_activator.c
 * Module:  Agora SD-RTN SDK RTC C API license activator.
 *
 * This is a tool to activate license Agora RTC Service SDK.
 * Copyright (C) 2020 Agora IO
 * All rights reserved.
 *
 *************************************************************/

#include <stdio.h>
#include <string.h>

#include <unistd.h>
#include <getopt.h>
#include <fcntl.h>

#include <arpa/inet.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <netdb.h>

// third-party
#include "jsmn.h"
#include "http_parser.h"

// Agora SDK
#include "AgoraBase.h"

#define LOGS(fmt, ...)  fprintf(stdout, "" fmt "\n", ##__VA_ARGS__)
#define LOGI(fmt, ...)  fprintf(stdout, "I/ " fmt " at %s:%u\n", ##__VA_ARGS__, __PRETTY_FUNCTION__, __LINE__)
#define LOGD(fmt, ...)  fprintf(stdout, "D/ " fmt " at %s:%u\n", ##__VA_ARGS__, __PRETTY_FUNCTION__, __LINE__)
#define LOGE(fmt, ...)  fprintf(stdout, "E/ " fmt " at %s:%u\n", ##__VA_ARGS__, __PRETTY_FUNCTION__, __LINE__)
#define LOGW(fmt, ...)  fprintf(stdout, "W/ " fmt " at %s:%u\n", ##__VA_ARGS__, __PRETTY_FUNCTION__, __LINE__)

#define TAG_APP     "[app]"

#define CERTIFACTE_FILENAME         "certificate.bin"
#define CREDENTIAL_FILENAME         "deviceID.bin"

#define INVALID_FD                  -1
#define MAX_TOKENS_COUNT            16
#define MAX_BUF_LEN                 1024
#define HTTP_MAX_BUF_LEN            4096

#define CREDENTIAL_MAX_LEN          256
#define CERTIFICATE_MAX_LEN         2048

typedef struct {
    const char  *p_certificate_dir;

    const char  *p_appid;
    const char  *p_customer_key;
    const char  *p_customer_secret;
    const char  *p_license_key;
} app_config_t;

typedef struct {
    app_config_t    config;

    jsmn_parser     json_parser_inst;
    http_parser     http_parser_inst;

    // output: save certificate & credential
    char            str_credential[CREDENTIAL_MAX_LEN];
    uint32_t        str_credential_len;
    char            str_certificate[CERTIFICATE_MAX_LEN];
    uint32_t        str_certificate_len;

    int32_t         b_activate_success;
    char            str_activate_result[MAX_BUF_LEN];
} app_t;

static app_t g_app_instance;

app_t *app_get_instance(void)
{
    static app_t *_app_ptr = NULL;
    if (!_app_ptr) {
        _app_ptr = &g_app_instance;
        _app_ptr->config.p_certificate_dir = ".";
        _app_ptr->config.p_appid = "";
        _app_ptr->config.p_customer_key = "";
        _app_ptr->config.p_customer_secret = "";
        _app_ptr->config.p_license_key = "";
        _app_ptr->str_credential_len = CREDENTIAL_MAX_LEN;
        _app_ptr->str_certificate_len = CERTIFICATE_MAX_LEN;
        _app_ptr->b_activate_success = 0;
    }
    return _app_ptr;
}

static int jsoneq(const char *json, jsmntok_t *tok, const char *s) {
  if (tok->type == JSMN_STRING && (int)strlen(s) == tok->end - tok->start &&
      strncmp(json + tok->start, s, tok->end - tok->start) == 0) {
    return 0;
  }
  return -1;
}

void app_print_usage(int32_t argc, char **argv)
{
    LOGS("\nUsage: %s [OPTION]", argv[0]);
    LOGS(" -h, --help               : show help info");
    LOGS(" -a, --appId              : application ID");
    LOGS(" -k, --customerKey        : customer key");
    LOGS(" -s, --customerSecret     : customer secret");
    LOGS(" -l, --licenseKey         : license key");
    LOGS(" -o, --certOutputDir      : certificate directory (output), default is '.'");
    LOGS("\nExample:");
    LOGS("    %s --appId xx --customerKey xx --customerSecret xx --licenseKey xx --certOutputDir .", argv[0]);
}

int32_t app_parse_args(app_config_t *p_config, int32_t argc, char **argv)
{
    const char* av_short_option = "ha:k:s:l:o:";
    const struct option av_long_option[] = {
        {"help",            0, NULL, 'h'},
        {"appId",           1, NULL, 'a'},
        {"customerKey",     1, NULL, 'k'},
        {"customerSecret",  1, NULL, 's'},
        {"licenseKey",      1, NULL, 'l'},
        {"certOutputDir",   1, NULL, 'o'},
        {0, 0, 0, 0}
    };

    int32_t ch = -1;
    int32_t optidx = 0;
    int32_t rval = 0;

    while (1) {
        optidx++;
        ch = getopt_long(argc, argv, av_short_option, av_long_option, NULL);
        if (ch == -1) {
            break;
        }

        switch (ch) {
        case 'h': {
            rval = -1;
            goto EXIT;
        } break;
        case 'a': {
            p_config->p_appid = optarg;
        } break;
        case 'k': {
            p_config->p_customer_key = optarg;
        } break;
        case 's': {
            p_config->p_customer_secret = optarg;
        } break;
        case 'l': {
            p_config->p_license_key = optarg;
        } break;
        case 'o': {
            p_config->p_certificate_dir = optarg;
        } break;
        default:{
            rval = -1;
            LOGS("%s parse cmd param: %s error.", TAG_APP, argv[optidx]);
            goto EXIT;
        }
        }
    }

    // check key parameters
    if (!p_config->p_appid || strcmp(p_config->p_appid, "") == 0) {
        rval = -1;
        LOGE("%s appid MUST be provided", TAG_APP);
        goto EXIT;
    }

    if (!p_config->p_customer_key || strcmp(p_config->p_customer_key, "") == 0) {
        rval = -1;
        LOGE("%s invalid customer key", TAG_APP);
        goto EXIT;
    }

    if (!p_config->p_customer_secret || strcmp(p_config->p_customer_secret, "") == 0) {
        rval = -1;
        LOGE("%s invalid customer secret", TAG_APP);
        goto EXIT;
    }

    if (!p_config->p_license_key || strcmp(p_config->p_license_key, "") == 0) {
        rval = -1;
        LOGE("%s invalid license key", TAG_APP);
        goto EXIT;
    }

    rval = access(p_config->p_certificate_dir, F_OK);
    if (rval != 0) {
        LOGE("%s directory doesn't exist", TAG_APP);
        goto EXIT;
    }

EXIT:
    return rval;
}

int32_t app_save_license_file(const char *path, char *buffer, uint32_t *buf_len)
{
    int32_t rval = 0;
    int32_t fd = 0;

    if (*buf_len <= 0) {
        LOGE("%s invalid buffer length:%d", TAG_APP, *buf_len);
        goto EXIT;
    }

    fd = open(path, O_CREAT|O_RDWR, S_IRUSR|S_IWUSR);
    if (fd == INVALID_FD) {
        rval = -1;
        LOGE("%s open file failed, path=%s", TAG_APP, path);
        goto EXIT;
    }

    *buf_len = write(fd, buffer, *buf_len);
    if (*buf_len <= 0) {
        rval = -1;
        LOGE("%s write file failed, path=%s rval=%d", TAG_APP, path, rval);
        goto EXIT;
    }

EXIT:
    if (fd > 0) {
        close(fd);
        fd = INVALID_FD;
    }
    return rval;
}

static int32_t _http_on_body(http_parser *p_parser, const char *at, size_t length)
{
    int32_t token_num = 0;
    int32_t str_len = 0;

    app_t *p_app = (app_t *)p_parser->data;

    jsmntok_t tokens[MAX_TOKENS_COUNT];
    jsmn_init(&p_app->json_parser_inst);
    int32_t rval = jsmn_parse(&p_app->json_parser_inst, at, length, tokens, MAX_TOKENS_COUNT);
    if (rval <= 0) {
        LOGE("%s parse config failed, token_num=%d", TAG_APP, rval);
        goto EXIT;
    }

    token_num = rval;
    if (token_num < 1 || tokens[0].type != JSMN_OBJECT) {
        rval = -1;
        LOGE("%s parse json failed", TAG_APP);
        goto EXIT;
    }

    for (int32_t i = 1; i < token_num; i++) {
        if (jsoneq(at, &tokens[i], "cert") == 0) {
            i++;
            str_len = tokens[i].end - tokens[i].start;
            memcpy(p_app->str_certificate, at + tokens[i].start, str_len);
            p_app->str_certificate[str_len] = '\0';
            p_app->str_certificate_len = str_len;
            p_app->b_activate_success = 1;
        } else {
            p_app->b_activate_success = 0;
            memcpy(p_app->str_activate_result, at, length);
            p_app->str_activate_result[length] = '\0';
            break;
        }
    }

EXIT:
    return 0;
}

int32_t base64_encode(const char *src, int32_t src_len, char *encoded)
{
    const char basis_64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

    int i;
    char *p = encoded;
    for (i = 0; i < src_len - 2; i += 3) {
        *p++ = basis_64[(src[i] >> 2) & 0x3F];
        *p++ = basis_64[((src[i] & 0x3) << 4) |
                        ((src[i + 1] & 0xF0) >> 4)];
        *p++ = basis_64[((src[i + 1] & 0xF) << 2) |
                        ((src[i + 2] & 0xC0) >> 6)];
        *p++ = basis_64[src[i + 2] & 0x3F];
    }
    if (i < src_len) {
        *p++ = basis_64[(src[i] >> 2) & 0x3F];
        if (i == (src_len - 1)) {
            *p++ = basis_64[((src[i] & 0x3) << 4)];
            *p++ = '=';
        } else {
            *p++ = basis_64[((src[i] & 0x3) << 4) |
                            ((src[i + 1] & 0xF0) >> 4)];
            *p++ = basis_64[((src[i + 1] & 0xF) << 2)];
        }
        *p++ = '=';
    }

    *p++ = '\0';
    return (p - encoded);
}

int32_t app_activate_license(app_t *p_app)
{
    // License Activation Flow
    // Reference Link
    // https://docs-preview.agoralab.co/cn/Agora%20Platform/license_mechanism_v3?platform=All%20Platforms

    char str_hostname[] = "api.agora.io";
    char str_ip[129] = "";
    int32_t sock_fd = 0;
    int32_t str_post_len = 0;
    int32_t str_authorization_base64_len = 0;
    int32_t http_buffer_len = 0;
    struct timeval timeout = { 0 };
    http_parser_settings http_activation_settings = { 0 };

    app_config_t *p_config = &p_app->config;
    struct addrinfo *p_addinfo = NULL;
    int32_t rval = 0;
    agora::util::AString astr_credential;
    rval = createAgoraCredential(astr_credential);
    if (rval != 0) {
        rval = -1;
        LOGE("generate credential failed, license_key=%s", p_config->p_license_key);
        LOGS("!!! please check whether your SDK supports license feature");
        goto EXIT;
    }
    p_app->str_credential_len = astr_credential.get()->length();
    memcpy(p_app->str_credential, astr_credential.get()->c_str(),
        (p_app->str_credential_len <= CREDENTIAL_MAX_LEN) ? p_app->str_credential_len : CREDENTIAL_MAX_LEN);
    LOGS("%s generate credential\nlen:%d\nstr:%s", TAG_APP, p_app->str_credential_len, p_app->str_credential);

    struct addrinfo hints;
    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_CANONNAME;
    hints.ai_protocol = 0;

    rval = getaddrinfo(str_hostname, NULL, &hints, &p_addinfo);
    if(rval != 0) {
        LOGE("%s dns parser failed", TAG_APP);
        goto EXIT;
    }

    for (struct addrinfo *res_p = p_addinfo; res_p != NULL; res_p = res_p->ai_next) {
        rval = getnameinfo(res_p->ai_addr, res_p->ai_addrlen, str_ip, sizeof(str_ip), NULL, 0, NI_NUMERICHOST);
        if(rval == 0) {
            break;
        }
    }

    if (strcmp(str_ip, "") == 0) {
        rval = -1;
        LOGE("%s dns parser can't get valid ip", TAG_APP);
        goto EXIT;
    }

    sock_fd = socket(AF_INET, SOCK_STREAM, 0);
    struct sockaddr_in serv_addr;
    memset(&serv_addr, 0, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(80);
    inet_pton(AF_INET, str_ip, &serv_addr.sin_addr);

    rval = connect(sock_fd, (struct sockaddr*) &serv_addr, sizeof(serv_addr));
    if (rval < 0) {
        LOGE("%s connect server failed, hostname:%s ip:%s", TAG_APP, str_hostname, str_ip);
        goto EXIT;
    }

    char str_page[MAX_BUF_LEN];
    snprintf(str_page, MAX_BUF_LEN, "/dev/v3/apps/%s/licenses/activation", p_config->p_appid);

    char str_post[MAX_BUF_LEN];
    str_post_len = snprintf(str_post, MAX_BUF_LEN,
                            "{ \"credential\": \"%s\", \"custom\": \"%s\" }",
                            p_app->str_credential, p_config->p_license_key);

    char str_authorization[MAX_BUF_LEN];
    char str_authorization_base64[MAX_BUF_LEN];
    snprintf(str_authorization, MAX_BUF_LEN, "%s:%s", p_config->p_customer_key, p_config->p_customer_secret);
    base64_encode(str_authorization, strlen(str_authorization), str_authorization_base64);

    char http_buffer[HTTP_MAX_BUF_LEN];
    http_buffer_len = snprintf(http_buffer, HTTP_MAX_BUF_LEN,
                            "POST %s HTTP/1.1\r\n"
                            "Authorization: Basic %s\r\n"
                            "Host: %s\r\n"
                            "Content-type: application/json\r\n"
                            "Accept: */*\r\n"
                            "Content-length: %d\r\n\r\n"
                            "%s",
                            str_page, str_authorization_base64, str_hostname, str_post_len, str_post);
    // LOGS("SEND len=%d\n%s", http_buffer_len, http_buffer);

    rval = write(sock_fd, http_buffer, http_buffer_len);
    if (rval <= 0) {
        rval = -1;
        LOGE("%s send license activate request failed", TAG_APP);
        goto EXIT;
    }

    fd_set rd_set;
    FD_ZERO(&rd_set);
    FD_SET(sock_fd, &rd_set);

    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    rval = select(sock_fd + 1, &rd_set, NULL, NULL, &timeout);
    if (rval == -1) {
        LOGE("%s wait http response timeout, rval=%d", TAG_APP, rval);
        goto EXIT;
    }

    rval = read(sock_fd, http_buffer, HTTP_MAX_BUF_LEN);
    if (rval <= 0) {
        rval = -1;
        LOGE("%s recv license activate request failed", TAG_APP);
        goto EXIT;
    }
    http_buffer_len = rval;
    http_buffer[http_buffer_len] = '\0';
    // LOGS("RECV len=%d\n%s", http_buffer_len, http_buffer);

    http_activation_settings.on_body = _http_on_body;

    p_app->http_parser_inst.data = (void*)p_app;
    http_parser_init(&p_app->http_parser_inst, HTTP_RESPONSE);
    http_parser_execute(&p_app->http_parser_inst, &http_activation_settings, http_buffer, http_buffer_len);
    if (!p_app->b_activate_success) {
        rval = -1;
        LOGE("%s activate failed", TAG_APP);
        LOGE("%s result: %s", TAG_APP, p_app->str_activate_result);
        goto EXIT;
    }
    LOGS("%s activate success", TAG_APP);

    char str_file_credential[MAX_BUF_LEN];
    snprintf(str_file_credential, MAX_BUF_LEN, "%s/%s", p_config->p_certificate_dir, CREDENTIAL_FILENAME);
    rval = app_save_license_file(str_file_credential, p_app->str_credential, &p_app->str_credential_len);
    if (rval != 0) {
        LOGE("%s save credential failed, path=%s", TAG_APP, str_file_credential);
        goto EXIT;
    }

    char str_file_certificate[MAX_BUF_LEN];
    snprintf(str_file_certificate, MAX_BUF_LEN, "%s/%s", p_config->p_certificate_dir, CERTIFACTE_FILENAME);
    rval = app_save_license_file(str_file_certificate, p_app->str_certificate, &p_app->str_certificate_len);
    if (rval != 0) {
        LOGE("%s save certificate failed, path=%s", TAG_APP, str_file_certificate);
        goto EXIT;
    }

    LOGS("---------------------  License Activation  --------------------");
    LOGS("|      %-54.52s |", "!!! PLEASE KEEP YOUR OWN UNIQUE CREDENTIAL");
    LOGS("|      %-54.52s |", "!!! IF CREDENTIAL IS LOST, IT CAN'T BE RETRIVED");
    LOGS("|      %-54.52s |", "");
    LOGS("|      %32.32s%-22.32s |", "Generated credential  is saved: ", str_file_credential);
    LOGS("|      %32.32s%-22.32s |", "Generated certificate is saved: ", str_file_certificate);
    LOGS("---------------------------------------------------------------");

EXIT:
    if (p_addinfo) {
        freeaddrinfo(p_addinfo);
    }

    if (sock_fd > 0) {
        close(sock_fd);
        sock_fd = INVALID_FD;
    }
    return rval;
}

int main(int argc, char **argv)
{
    app_t *p_app = app_get_instance();

    // 0. app parse args
    int32_t rval = app_parse_args(&p_app->config, argc, argv);
    if (rval != 0) {
        app_print_usage(argc, argv);
        goto EXIT;
    }

    // 1. license activate
    app_activate_license(p_app);

EXIT:
    return rval;
}

